# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-08-24 12:14
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
from django.db.models import Q
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('users', '0002_auto_20170823_1007'),
    ]

    def forward_func(apps, schema_editor):
        Team = apps.get_model("users", "Team")
        TeamMembership = apps.get_model("users", "TeamMembership")

        db_alias = schema_editor.connection.alias

        # Create team membership element for each member
        memberships = []
        for team in Team.objects.using(db_alias).all():
            members = team.members.user_set.all()
            admins = team.admins.user_set.all()
            for member in members:
                memberships.append(TeamMembership(
                    team=team,
                    user=member,
                    is_admin=member in admins))

        TeamMembership.objects.using(db_alias).bulk_create(memberships)

    def backward_func(apps, schema_editor):
        TeamMembership = apps.get_model("users", "TeamMembership")
        db_alias = schema_editor.connection.alias
        TeamMembership.objects.using(db_alias).all().delete()

    def delete_group_forward_func(apps, schema_editor):
        Group = apps.get_model("auth", "Group")
        db_alias = schema_editor.connection.alias

        Group.objects.using(db_alias).filter(
            Q(name__endswith='_members') | Q(name__endswith='_admins')).delete()

    def delete_group_backward_func(apps, schema_editor):
        raise NotImplementedError

    operations = [
        migrations.CreateModel(
            name='TeamMembership',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_admin', models.BooleanField(default=False)),
            ],
        ),
        migrations.AddField(
            model_name='teammembership',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='users.Team'),
        ),
        migrations.AddField(
            model_name='teammembership',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='team_memberships', to=settings.AUTH_USER_MODEL),
        ),
        migrations.RunPython(forward_func, backward_func, atomic=True),
        migrations.RemoveField(
            model_name='team',
            name='admins',
        ),
        migrations.RemoveField(
            model_name='team',
            name='members',
        ),
        migrations.AddField(
            model_name='team',
            name='members',
            field=models.ManyToManyField(through='users.TeamMembership', to=settings.AUTH_USER_MODEL),
        ),
        migrations.RunPython(delete_group_forward_func, delete_group_backward_func, atomic=True),
    ]
